from math import sqrt, pow



def integral(self, e_down, e_up, inter="const"):
	# INPUT CHECK
	if (e_down > e_up):
		raise ValueError("\'e_down\' is larger than \'e_up\'")
	if (e_down < 0):
		raise ValueError("\'e_down\' cannot be negaitve")
	if (e_up < 0):
		raise ValueError("\'e_up\' cannot be negaitve")
	if (e_down < self._e_low[0]):
		raise ValueError(
			"\'e_down\' is lower than lowest possible energy = {:f}".format(
				self._e_low[0]
			)
		)
	if (e_up > self._e_high[-1]):
		raise ValueError(
			"\'e_down\' is higher than highest possible energy = {:f}".format(
				self._e_high[-1]
			)
		)
	if (e_down == e_up):
		raise ValueError(
			"\'e_down\' is the same as \'e_up\' with value = {:f}".format(
				e_down
			)
		)

	# GROUPS WITHIN INTERVAL (E_DOWN, E_UP)
	groups = self._groups_in_interval(e_down, e_up)

	# FLUX IN LOWEST GROUP OR ITS FRACTION
	if (self._e_low[groups[0]-1] == e_down):
		delta_e    = self._e_high[groups[0]-1] - self._e_low[groups[0]-1]
		lower_flux = self._flux[groups[0]-1] * delta_e
	else:
		group_width = self._e_high[groups[0]-1] - self._e_low[groups[0]-1]
		delta_e     = self._e_high[groups[0]-1] - e_down
		lower_flux  = self._flux[groups[0]-1] * delta_e

	# FLUX IN LOWEST GROUP OR ITS FRACTION
	if (self._e_high[groups[-1]-1] == e_up):
		delta_e    = self._e_high[groups[-1]-1] - self._e_low[groups[-1]-1]
		upper_flux = self._flux[groups[-1]-1] * delta_e
	else:
		group_width = self._e_high[groups[0]-1] - self._e_low[groups[0]-1]
		delta_e     = e_up-self._e_low[groups[-1]-1]
		upper_flux  = self._flux[groups[-1]-1] * delta_e

	# INTEGRAL
	integral = lower_flux
	for group in groups[1:-1]:
		delta_e   = self._e_high[group-1] - self._e_low[group-1]
		integral += self._flux[group-1] * delta_e
	if (len(groups)>1):
		integral += upper_flux
	else:
		delta_e  = self._e_high[groups[0]-1] - self._e_low[groups[0]-1]
		integral = self._flux[groups[0]-1] * delta_e
		integral = lower_flux + upper_flux - integral

	# INTEGRAL UNCERTAINTY
	uncertainty = 0
	for group in groups:
		uncertainty += pow(self._stdev[group-1], 2)
	uncertainty = sqrt(uncertainty)

	return [e_down, e_up, integral, uncertainty]



def _integrity_check(self):
	for group in range(1, len(self._grp), 1):
		if not(self._e_high[group-1] == self._e_low[group]):
			raise ValueError(
				"\'e_up\' of energy group {:d} and ".format(group) +
				"\'e_down\' of energy group " +
				"{:d}are not identical".format(group+1)
			)
		if not(self._grp[group-1] == self._grp[group]-1):
			raise ValueError(
				"energy group {:d} is missing".format(group)
			)

	for group in self._grp:
		if not(self._flux[group-1] >= 0):
			raise ValueError(
				"flux in energy group {:d} is negative".format(group)
			)
		if not(self._flux[group-1] >= 0):
			raise ValueError(
				"uncertainty in energy group {:d} is negative".format(group)
			)



import json



def dump_txt(self, path):
	with open(path, "w") as fs:
		fs.write("#{:} {:11} {:11} {:11} {:5}\n".format(
			"grp",
			"e_down",
			"e_up",
			"flux",
			"stdev"
		))
		for group in self._grp:
			fs.write("{:04d} {:.5E} {:.5E} {:.5E} {:06.3f}\n".format(
				self._grp[group-1],
				self._e_low[group-1],
				self._e_high[group-1],
				self._flux[group-1],
				self._stdev[group-1]
			))

def dump_json(self, path):
	dict = {}
	for group in self._grp:
		dict[str(group)] = {}
		dict[str(group)]["e_down"] = self._e_low[group-1]
		dict[str(group)]["e_up"]   = self._e_high[group-1]
		dict[str(group)]["flux"]   = self._flux[group-1]
		dict[str(group)]["stdev"]  = self._stdev[group-1]
	with open(path, "w") as fs:
		json.dump(fs, dict)



def __eq__(self, other):
	# LOW ENERGIES COMPARISON
	if not(len(self._e_low) == len(other._e_low)):
		return False
	else:
		for group in range(1, len(self._e_low)+1, 1):
			if not(self._e_low[group-1] == other._e_low[group-1]):
				return False
	# HIGH ENERGIES COMPARISON
	if not(len(self._e_high) == len(other._e_high)):
		return False
	else:
		for group in range(1, len(self._e_high)+1, 1):
			if not(self._e_high[group-1] == other._e_high[group-1]):
				return False
	# FLUXES COMPARISON
	if not(len(self._flux) == len(other._flux)):
		return False
	else:
		for i in range(1, len(self._flux)+1, 1):
			if not(self._flux[group-1] == other._flux[group-1]):
				return False
	# UNCERTAINTIES COMPARISON
	if not(len(self._stdev) == len(other._stdev)):
		return False
	else:
		for i in range(1, len(self._stdev)+1, 1):
			if not(self._stdev[group-1] == other._stdev[group-1]):
				return False

	return True



def regroup(self, e_low, e_high, inter="const"):
	tmp_grp    = []
	tmp_e_low  = []
	tmp_e_high = []
	tmp_flux   = []
	tmp_stdev  = []
	for group in range(1,len(e_low)+1,1):
		# WITHIN DEFINED ENERGIES OF SPECTRUM BEEING REGROUPED
		if (
			(e_low[group-1]  >= self._e_low[0])   and
			(e_low[group-1]  <= self._e_high[-1]) and
			(e_high[group-1] >= self._e_low[0])   and
			(e_high[group-1] <= self._e_high[-1])
		):
			integral = self.integral(e_low[group-1], e_high[group-1])
			tmp_grp.append(group)
			tmp_e_low.append(e_low[group-1])
			tmp_e_high.append(e_high[group-1])
			tmp_flux.append(integral[2] / (e_high[group-1] - e_low[group-1]))
			tmp_stdev.append(integral[3])

		# OUTSIDE DEFINED ENERGIES OF SPECTRUM BEEING REGROUPED
		if (
			(e_low[group-1]  <  self._e_low[0])   and
			(e_high[group-1] <= self._e_low[0])   or
			(e_low[group-1]  >= self._e_high[-1]) and
			(e_high[group-1] >  self._e_high[-1])
		):
			tmp_grp.append(group)
			tmp_e_low.append(e_low[group-1])
			tmp_e_high.append(e_high[group-1])
			tmp_flux.append(0.0)
			tmp_stdev.append(0.0)

		# PARTIALY DEFINED ENERGIES OF SPECTRUM BEEING REGROUPED
		# LOWER END
		if (
			(e_low[group-1]  <  self._e_low[0])   and
			(e_high[group-1] > self._e_low[0])   and
			(e_high[group-1] <= self._e_high[-1])
		):
			integral = self.integral(self._e_low[0], e_high[group-1])
			tmp_grp.append(group)
			tmp_e_low.append(e_low[group-1])
			tmp_e_high.append(e_high[group-1])
			tmp_flux.append(integral[2] / (e_high[group-1] - self._e_low[0]))
			tmp_stdev.append(integral[3])

		# PARTIALY DEFINED ENERGIES OF SPECTRUM BEEING REGROUPED
		# UPPER END
		if (
			(e_low[group-1]  >= self._e_low[0])   and
			(e_low[group-1]  <= self._e_high[-1]) and
			(e_high[group-1] >  self._e_high[-1])
		):
			integral = self.integral(e_low[group-1], self._e_high[-1])
			tmp_grp.append(group)
			tmp_e_low.append(e_low[group-1])
			tmp_e_high.append(e_high[group-1])
			tmp_flux.append(integral[2] / (self._e_high[-1] - e_low[group-1]))
			tmp_stdev.append(integral[3])

		# PARTIALY DEFINED ENERGIES OF SPECTRUM BEEING REGROUPED
		# BOTH ENDS
		if (
			(e_low[group-1]  < self._e_low[0])   and
			(e_high[group-1] > self._e_high[-1])
		):
			integral = self.integral(self._e_low[0], self._e_high[-1])
			tmp_grp.append(group)
			tmp_e_low.append(e_low[group-1])
			tmp_e_high.append(e_high[group-1])
			tmp_flux.append(integral[2] / (self._e_high[-1] - self._e_low[0]))
			tmp_stdev.append(integral[3])

	return self.__class__(
			source = [tmp_grp, tmp_e_low, tmp_e_high, tmp_flux, tmp_stdev]
		)



def _groups_in_interval(self, e_down, e_up):
	groups = []
	for grp in self._grp:
		if (self._e_low[grp-1] >= e_down and self._e_high[grp-1] <= e_up):
			if not(grp in groups):
				groups.append(grp)
		else:
			if (self._e_low[grp-1] < e_down and self._e_high[grp-1] > e_down):
				if not(grp in groups):
					groups.append(grp)
			if (self._e_low[grp-1] < e_up and self._e_high[grp-1] > e_up):
				if not(grp in groups):
					groups.append(grp)
	return groups

def group(self, number):
	return([
		self._grp[number-1],
		self._e_low[number-1],
		self._e_high[number-1],
		self._flux[number-1],
		self._stdev[number-1]
	])



from io import IOBase



def __init__(self, source=None, name=None, group_flux="diff"):
	if not(group_flux in ["diff", "intg"]):
		raise AttributeError(
			"\'group_flux\' attribute takes values \'diff\' or \'intg\'"
		)
	if not(
		isinstance(source, IOBase)      or \
		isinstance(source, list)        or \
		isinstance(source, type(None))
	):
		raise AttributeError(
			"\'source\' attribute must be of class \'list\' or \'IOBase\'"
		)

	self._grp    = []
	self._e_low  = []
	self._e_high = []
	self._flux   = []
	self._stdev  = []
	if (isinstance(source, IOBase)):
		for line in source:
			if not("#" == line[0]):
				self._grp.append(int(line.split()[0]))
				self._e_low.append(float(line.split()[1]))
				self._e_high.append(float(line.split()[2]))
				self._flux.append(float(line.split()[3]))
				self._stdev.append(float(line.split()[4]))
	if (isinstance(source, list)):
		self._grp    = source[0]
		self._e_low  = source[1]
		self._e_high = source[2]
		self._flux   = source[3]
		self._stdev    = source[4]
	self._name = name

	if (group_flux=="intg"):
		for group in self._grp:
			delta_e = self._e_high[group-1] - self._e_low[group-1]
			self._flux[group-1] = self._flux[group-1] / delta_e

	self._integrity_check()



class NSpectrum:
	_name       = None
	_grp        = []
	_e_low      = []
	_e_high     = []
	_flux       = []
	_stdev      = []

	from .spectrum_methods._constructor  import __init__
	from .spectrum_methods._integrity    import _integrity_check
	from .spectrum_methods._equal        import __eq__
	from .spectrum_methods._integral     import integral
	from .spectrum_methods._groups       import _groups_in_interval
	from .spectrum_methods._groups       import group
	from .spectrum_methods._regroup      import regroup
	from .spectrum_methods._export       import dump_txt
	from .spectrum_methods._export       import dump_json

	def max_energy(self):
		return self._e_high[-1]

	def min_energy(self):
		return self._e_low[0]

	def groups(self):
		return self._grp

	def es_low(self):
		return self._e_low

	def es_high(self):
		return self._e_high

	def fluxes(self):
		return self._flux

	def stdevs(self):
		return self._stdev

	def name(self):
		return self._name
